#!/usr/bin/python3
import serial
import os

H = 1
L = 0
FUN_CODE_DIG_READ = 3
FUN_CODE_DIG_WRITE = 4
FUN_CODE_AN_READ = 5
FUN_CODE_AN_WRITE = 6
FUN_CODE_GET_IDS = 8 #maps to FUN_CODE_TX_IDs in the arduino code
FUN_CODE_CONFIG = 2
BROADCAST_ADDR = 255

DEBUG = False

class Uno:
    '''Arduino Uno as an IO slave.
    
    The Uno must be running the Arduino_slavexxx sketch.
    Uses an ascii formatted communications telegram.
    
    '''
    
    def __init__(self,address = 0, project = 0, config_digital = []):
        '''Initialise the Uno slave.
        
        address -- 1 - 254.
        project -- 0 - 65535
        config_digital -- List of integers detailing the pins configured as outputs.
        
        '''
        
        self.project = project
        self.address = address
        self.config_digital = config_digital
        self.port = None
        self.rw_available__flag = False
        self.ser = serial.Serial()             
        self.ser.timeout = 0.1
        self.ser.bytesize = 8
        self.ser.stopbits = 2
        self.ser.baudrate = 9600
        self.ser.xonxoff = False
        self.write_digital_dict = {}
        self.write_analog_dict = {}
        self.read_digital_dict = {0:0, 1:0, 2:0, 3:0, 4:0, 5:0, 6:0, 7:0, 8:0, 9:0, 10:0, 11:0, 12:0, 13:0, }
        self.read_analog_dict = {0:0, 1:0, 2:0, 3:0, 4:0, 5:0}

    def connection(self):
        '''This function checks for connection and digital configuration and calls connect() if neccessary.
        
        Is called by read_digital(), read_analog(), write_digital(), and write_analog() to check for connection to slave.
        Return True or False.
        
        '''
        
        tries = 0
        while (self.ser.isOpen() == False or self.rw_available__flag == False) and tries < 5 :      
            print('tries {0:d}'.format(tries))
            tries += 1        
            if self.connect():
                if self.config_digital_io():
                    self.rw_available__flag = True
                    
                else:
                    print('Failed to configure digital outputs')
            else:
                print('Failed to connect')
            
                    
        if self.rw_available__flag:
            return True         
        else:
            return False   
                          
    def connect(self):
        '''Called by self.connection() to find the slave and its port.
        
        Opens the serial port and configures the digital outputs by calling config_digital_io().
        Return True or False'
        
        '''
        
        acms = []
        file_list = os.listdir('/dev')
        for fl in file_list:
            if fl[3:6] == 'ACM':
                acms.append(fl)
                
        for acm in acms:
            try:
                if self.ser.isOpen():
                    self.ser.close()
                
                self.ser.port = '/dev/'+ acm
                self.ser.open()
                telegram_bin = bytearray()
                telegram_bin.append(BROADCAST_ADDR)
                telegram_bin.append(FUN_CODE_GET_IDS)
                fcs= self.check_sum(telegram_bin)
                telegram_bin.append(fcs)
                
                telegram_ascii_tx = b':' + self.bin_to_ascii(telegram_bin) + b'\n'
                self.ser.write(telegram_ascii_tx)
                if DEBUG: print('In connect telegram_ascii_tx', telegram_ascii_tx)
                telegram_received_binary = self.telegram_read()
                if telegram_received_binary == None:
                    return  False                                  
                chk_sum = self.check_sum(telegram_received_binary[:-1])
                
                project_received = (telegram_received_binary[2] << 8) + telegram_received_binary[3]
                if self.address == telegram_received_binary[0] and self.project == project_received:
                    self.port = acm
                    return True
                                         
            except Exception as e:
                print('In connect', e)
                return False
                   

        print('Incorrect address or project number for slave, project = {0:d} address = {1:d}'.format(self.project, self.address)) 
        return False




    def check_sum(self,frame):
        '''Calculates and returns the check sum for a telegram. 
        
        frame -- A bytearray to be have its check sum calculated, note if it is a received telegram the caller must strip off the
                 received check sum first.      
        Return  check sum integer.
        
        '''
  
        chk_sum = 0;
        for n in frame:
            chk_sum = chk_sum + n
            
        return chk_sum%256

    def bin_to_ascii(self,telegram_bin):
        '''Converts a binary format telegram to an ASCII format telegram. 
        
        telegram_bin -- bytearray in binary format.
            
        Return bytearray of telegram in ascii format.
        
        '''
        
        telegram_ascii = bytearray()
        bin2ascii = {0:48,1:49,2:50,3:51,4:52,5:53,6:54,7:55,8:56,9:57,10:65,11:66,12:67,13:68,14:69,15:70}    
        for byte in telegram_bin:
                telegram_ascii.append(bin2ascii[byte >> 4] )
                telegram_ascii.append(bin2ascii[byte & 0x0F])
        return telegram_ascii

    def ascii_to_bin(self,telegram_ascii):
        '''Converts an ASCII format telegram to a binary format telegram. 
        
        telegram_ascii -- bytearray telegram in ascii format.
        Return bytearray in binary format.
        
        '''
        
        ascii2bin = {48:0,49:1,50:2,51:3,52:4,53:5,54:6,55:7,56:8,57:9,65:10,66:11,67:12,68:13,69:14,70:15}
        telegram_bin = bytearray()
        
        while telegram_ascii:
            a = telegram_ascii.pop(0)
            a = ascii2bin[a] << 4
            b = telegram_ascii.pop(0)
            b = ascii2bin[b] 
            telegram_bin.append(a + b)

        return telegram_bin    

    def telegram_read(self):
        '''Receives a telegram and checks it for errors. 
            
        Return bytearray of telegram received from serial port will be in ascii format.
            
        ''' 
            
        try:
            rc_ch = 0
            telegram_string = b''
            bytecount = 0
            while (not rc_ch == b'\n') and bytecount < 62:
                rc_ch = self.ser.read()
                telegram_string = telegram_string + rc_ch
                bytecount+=1
            if bytecount > 61:
                return b''
                    
            telegram_received_ascii  = bytearray(telegram_string)
            if DEBUG: print('In telegram read telegram_received_ascii', telegram_received_ascii)
            if not telegram_received_ascii[0] == 58:
                return b''
            
            telegram_received_ascii.pop(0)
            telegram_received_ascii.pop()
            
            telegram_received_binary = self.ascii_to_bin(telegram_received_ascii)
            chksum_received = telegram_received_binary[-1]
            chksum = self.check_sum(telegram_received_binary[:-1])
            if chksum != chksum_received:
                print('Check sum error')
            else:
                if telegram_received_binary[1] != 0x80:              #Checks that there is no error report from the slave
                    return telegram_received_binary                  #Returns the healthy telegram in binary format.
                elif telegram_received_binary[2] == 1:               #Decodes the error report
                    print('Check sum error report from the slave')
                
                elif telegram_received_binary[2] == 2:
                    print('Function code error report from the slave')
                
                elif telegram_received_binary[2] == 3:
                    print('Address error report from the slave')
                
                elif telegram_received_binary[2] == 4:
                    print('Project number error report from the slave')
                
                else:
                    print('Unknown error')
            
            return b''
                                
        except Exception as e:
            print('In telegram_read() ', e)
            return b''
            
    def config_digital_io(self):
        '''Configures the digital IO. 
        
        Called by connect()
        Returns True or False.
        
        '''
        
        try:
            l = len(self.config_digital)
            telegram_bin = bytearray()
            telegram_bin.append(self.address)
            telegram_bin.append(FUN_CODE_CONFIG)
            telegram_bin.append(self.project >> 8)
            telegram_bin.append(self.project & 0xFF)
            config_portD = 0
            config_portB = 0
            for config in self.config_digital:
                if config < 8:
                    config_portD = config_portD | (1 << config ) 
                else:
                    config_portB = config_portB | (1 << (config - 8)) 
            telegram_bin.append(config_portD)
            telegram_bin.append(config_portB)
            fcs= self.check_sum(telegram_bin)
            telegram_bin.append(fcs)
            telegram_ascii_tx = b':' + self.bin_to_ascii(telegram_bin) + b'\n'
            self.ser.write(telegram_ascii_tx)
            if DEBUG: print('In config_digital telegram_ascii_tx', telegram_ascii_tx)
            telegram_received_binary = self.telegram_read()
            if telegram_received_binary == None:
                return False
            else:
                return True
                
        except Exception as e:
            return False
            
    def read_digital(self):
        '''Obtains the state of the digital inputs from the slave.
        
        Values are placed in the self.read_digtal_dict 
        Return a dictionary of digital input values.
        
        '''
        
        if self.connection():
            try:
                telegram_bin = bytearray()
                telegram_bin.append(self.address)
                telegram_bin.append(FUN_CODE_DIG_READ)
                fcs = self.check_sum(telegram_bin)
                telegram_bin.append(fcs)
                telegram_ascii_tx = b':' + self.bin_to_ascii(telegram_bin) + b'\n' 
                self.ser.write(telegram_ascii_tx)
                telegram_binary = self.telegram_read()
                return_data_dict = {}
                for pin in range (2,14):
                    if pin < 8:
                        a =  ((telegram_binary[2] & (1 << pin) )) >> pin
                    else:
                        a = (telegram_binary[3] & (1 << (pin -8) )) >> (pin -8)
                    self.read_digital_dict[pin] = a

                return True
            except:
                self.rw_available__flag = False
                return  False           
        else:
            return False

    def read_analog(self,number_of_pins):
        '''Reads the values from the analog inputs 0 - 5 depending on how many are asked for. 
        
        number_of_pins -- integer 1 - 6 number of consecutive pins read starting from A0.
        Values are placed in the self.read_digtal_dict. Values 0 - 1023 for 0 - 5V.
        
        '''
        
        if self.connection():
            try:
                telegram_bin = bytearray()
                telegram_bin.append(self.address)
                telegram_bin.append(FUN_CODE_AN_READ)
                telegram_bin.append(number_of_pins)
                fcs = self.check_sum(telegram_bin)
                telegram_bin.append(fcs)
                telegram_ascii_tx = b':' + self.bin_to_ascii(telegram_bin) + b'\n'

                self.ser.write(telegram_ascii_tx)
                
                telegram_binary = self.telegram_read()
                if telegram_binary != b'':
                    for n in range (0,number_of_pins*2, 2):
                        self.read_analog_dict[n//2] = (telegram_binary[n+2] << 8) + telegram_binary[n+3]

                return True
                
            except:
                self.rw_available__flag = False
                return True     
     
        else:
            return False
                         
    def write_digital(self):
        '''Writes to the digital outputs in the slave.
        
        Writes the self.write_digital_dict to the pins
        Return True or False.
        
        '''
        
        if self.connection():
            try:  
                telegram_bin = bytearray()
                telegram_bin.append(self.address)
                telegram_bin.append(FUN_CODE_DIG_WRITE)
                data_portD = 0
                data_portB = 0
                for key in self.write_digital_dict:
                    if key < 8:
                        data_portD = data_portD | (self.write_digital_dict[key] << key) 
                    else:
                        data_portB = data_portB | (self.write_digital_dict[key] << (key-8))
                telegram_bin.append(data_portD)
                telegram_bin.append(data_portB)
                fcs = self.check_sum(telegram_bin)
                telegram_bin.append(fcs)
                telegram_ascii_tx = b':' + self.bin_to_ascii(telegram_bin) + b'\n'
                self.ser.write(telegram_ascii_tx) 
                telegram_binary = self.telegram_read()
                if telegram_binary != b'':
                    return True
                else:
                    return False
            except:
                self.rw_available__flag = False
                return False

        else:
            return False

    def write_analog(self):
        '''Writes a value between 0 - 255 to the pulse width modulated pins 
        
        Writes the self.write_analog_dict to the pins
        Return True or False.
        
        '''
        
        if self.connection():
            try:
                telegram_bin = bytearray()
                telegram_bin.append(self.address)
                telegram_bin.append(FUN_CODE_AN_WRITE)
                for key in self.write_analog_dict:
                    telegram_bin.append(key & 0xFF)
                    telegram_bin.append(self.write_analog_dict[key] & 0xFF)
                    
                
                fcs = self.check_sum(telegram_bin)
                telegram_bin.append(fcs)
                telegram_ascii_tx = b':' + self.bin_to_ascii(telegram_bin) + b'\n'
                self.ser.write(telegram_ascii_tx)
                telegram_binary = self.telegram_read()
                if telegram_binary != b'':
                    return True
                else:
                    return False
                    
            except:
                self.rw_available__flag = False
                return False

        else:
            return False

