#!/usr/bin/env python3
#curses_widgets.py

import curses

class IndicatorLamp:
    def __init__(self, std_scr, start_y = 0, start_x = 0, captions = ['   ', '   '], colorpair_off = 1, colorpair_on = 2):
        self.stdscr = std_scr
        width = len(max(captions, key = len)) + 2 
        self.x_dict = {0:width//2 - len(captions[0])//2, 1:width//2 - len(captions[1])//2}
        self.captions = captions
        self.colorpair_off = colorpair_off
        self.colorpair_on  = colorpair_on      
        self.win = curses.newwin(3, width, start_y, start_x)
        self.set_state(False)        

    def set_state(self, state):
        self.win.clear()
        if state:
            self.win.bkgd(' ', curses.color_pair(self.colorpair_on))
        else:            
            self.win.bkgd(' ', curses.color_pair(self.colorpair_off))

        self.win.addstr(1, self.x_dict[int(state)], self.captions[int(state)], curses.A_BOLD)

class AnalogGauge:
    def __init__(self, std_scr, start_y = 0, start_x = 0, fullscale = 1, units = 'Units', colorpair = 0):
        self.stdscr = std_scr        
        self.scale = fullscale
        self.win = curses.newwin(5, 102, start_y, start_x)
        self.win.bkgd(' ', curses.color_pair(colorpair) | curses.A_BOLD)
        for n in range (0, 101):
            if n %10 == 0:
                self.win.addstr(1,  n, "|")

        self.win.addstr(1, 3, units)
        self.win.box()
        self.drive(0)        

    def drive(self, value):
        deflection = 100*value/self.scale
        if deflection >  100:
            deflection = 100.0     
        
        self.win.addstr(3, 3, '{0:06.2f}'.format(value))
        self.win.clrtoeol() 
        deflection = int(deflection)
        self.win.addstr(2 , 1, '|'*deflection)
        self.win.clrtoeol()
        self.win.box()


class ThumbWheel:  
    def __init__(self, std_scr, start_y = 0, start_x = 0, colorpair = 0):
        curses.mousemask(curses.BUTTON1_PRESSED)
        self.stdscr = std_scr        
        self.start_x = start_x
        self.start_y = start_y  
        self.win = curses.newwin(3, 11, start_y , start_x)
        self.win.bkgd(' ', curses.color_pair(colorpair) | curses.A_BOLD)
        self.win.box()
        self.digit = {2:0, 3:0, 4:0, 5:0, 7:0, 8:0, 9:0} 
        self.sign = 1
        self.value = 0.0
        self.update_wheels()

    def inc_dec(self,x, y):
        if x > 1 and x < 10 and x != 6:
            if y == 0:
                self.digit[x] = self.digit[x] + 1           

            if y == 2:
                self.digit[x] = self.digit[x] - 1

            if self.digit[x] > 9: 
                self.digit[x] = 0

            if self.digit[x] < 0:
                self.digit[x] = 9 

    def poll(self, mouseevent):
        _, x, y, _, event_type = mouseevent
        if event_type == curses.BUTTON1_PRESSED:
            x = x - self.start_x
            y = y - self.start_y 

            if x == 1 and y >= 0 and y <= 2:
               self.sign = -1 * self.sign
            else:
                self.inc_dec(x, y)

            self.update_wheels()

            self.value = self.sign * (1000*self.digit[2] + 100*self.digit[3] + 10*self.digit[4] + self.digit[5] + self.digit[7]/10 + 
                              self.digit[8]/100 + self.digit[9]/1000)

    def write(self, new_value):
        self.value = new_value
        if new_value >= 0:
            s = '+{0:08.3f}'.format(new_value)
            self.sign = 1
        else:
            s = '{0:09.3f}'.format(new_value)
            self.sign = -1
    
        self.digit[2] = int(s[1])
        self.digit[3] = int(s[2])
        self.digit[4] = int(s[3])
        self.digit[5] = int(s[4])
        self.digit[7] = int(s[6])
        self.digit[8] = int(s[7])
        self.digit[9] = int(s[8])
        self.update_wheels()

    def update_wheels(self):
            if self.sign == 1:
                sign_string = '+'
            else:
                sign_string = '-'

            self.win.addstr(1, 1, '{0:s}{1:d}{2:d}{3:d}{4:d}.{5:d}{6:d}{7:d}'.format(sign_string, self.digit[2],self.digit[3],
                                                                                      self.digit[4],self.digit[5],self.digit[7],
                                                                                      self.digit[8],self.digit[9]))
           
class Button:  
    def __init__(self, std_scr, start_y = 0, start_x = 0, caption = 'Button', colorpair = 0):
        curses.mousemask(curses.BUTTON1_PRESSED)
        self.stdscr = std_scr        
        self.start_x = start_x
        self.start_y = start_y 
        self.width = len(caption) + 2
        self.win = curses.newwin(3, self.width, start_y , start_x)
        self.win.bkgd(' ', curses.color_pair(colorpair) | curses.A_BOLD)
        self.win.box()
        self.win.addstr(1, 1, caption)
        self.state = False
        self.stdscr.refresh()
        self.win.refresh()

    def poll(self, mouseevent):
        _, x, y, _, event_type = mouseevent
        if event_type == curses.BUTTON1_PRESSED:
            x = x - self.start_x
            y = y - self.start_y
 
        if x > 0 and x < self.width and    y >= 0 and y <= 2:
            self.state = True  


class Label:  
    def __init__(self, std_scr, width = 8, start_y = 0, start_x = 0, colorpair = 0):
        self.stdscr = std_scr
        self.start_y = start_y 
        self.start_x = start_x 
        self.width = width
        self.colorpair = colorpair
        self.win = curses.newwin(3, width, start_y , start_x)
        self.win.bkgd(' ', curses.color_pair(colorpair) | curses.A_BOLD)        
        self.write('xxxxxxxx')       

    def write(self, caption):
        self.win.addstr(1, 1, caption[:self.width-2])
        self.win.clrtoeol()
        self.win.box()


if __name__ == '__main__':
    def main(stdscr):
        curses.curs_set(0)
        stdscr.timeout(0)
        curses.init_pair(1, curses.COLOR_WHITE, curses.COLOR_RED)
        curses.init_pair(2, curses.COLOR_WHITE, curses.COLOR_GREEN)
        curses.init_pair(3, curses.COLOR_WHITE, curses.COLOR_BLUE)

        indicator1 = IndicatorLamp(stdscr, start_y = 10, start_x = 5, captions = ['Ind1_off','Ind1_on'], colorpair_on = 3)
        indicator2 = IndicatorLamp(stdscr, start_y = 14, start_x = 5, captions = ['Ind2_off','Ind2_on'])
        
        gauge1 = AnalogGauge(stdscr, start_y = 20, start_x = 5, fullscale = 50, units = 'Volts')    
        gauge2 = AnalogGauge(stdscr, start_y = 25, start_x = 5, fullscale = 200, colorpair = 3)    

        twheel1 = ThumbWheel(stdscr, start_y = 10, start_x = 20, colorpair = 3)
        twheel2 = ThumbWheel(stdscr, start_y = 14, start_x = 20)
        twheel1.write(1.23456789)
        
        button1 = Button(stdscr, start_y = 1 , start_x = 5, caption = '   No1   ', colorpair = 3)
        button2 = Button(stdscr, start_y = 5 , start_x = 5, caption = '   No2   ')

        label1 = Label(stdscr, width = 16, start_y = 10, start_x = 40, colorpair = 3)
        label2 = Label(stdscr, width = 16, start_y = 14, start_x = 40)


        while True:
            event = stdscr.getch()
            if event == ord('q'):
                return
         
            
            elif event == curses.KEY_MOUSE:
                try:
                    mouse_event = curses.getmouse()
                    twheel1.poll(mouse_event)
                    twheel2.poll(mouse_event)
                    button1.poll(mouse_event)
                    button2.poll(mouse_event)
                except Exception as e:
                    stdscr.addstr(35,10, str(e), curses.A_BOLD)
                  
            if button1.state:
                indicator1.set_state(True)
                indicator2.set_state(False)                   
                gauge1.drive(20.02)
                gauge2.drive(50.0)
                button1.state = False                   

            if button2.state:
                indicator2.set_state(True)
                indicator1.set_state(False)                   
                gauge2.drive(20.02)
                gauge1.drive(24.57)
                button2.state = False

            label1.write(str(twheel1.value)+ ' Volts')
            label2.write(str(twheel2.value)+ ' Amps') 
                    
            stdscr.refresh()
            indicator1.win.refresh()           
            indicator2.win.refresh()           

            gauge1.win.refresh()           
            gauge2.win.refresh()           

            twheel1.win.refresh()           
            twheel2.win.refresh()           

            label1.win.refresh()           
            label2.win.refresh()           


            curses.napms(100)

    curses.wrapper(main)

